using System;
using System.Collections;
using System.ComponentModel;
using System.Drawing;
using System.Windows.Forms;

using Microsoft.DirectX;
using Microsoft.DirectX.Direct3D;

namespace zBufferDemo
{
	public class MainClass : Form
	{
		private const int numberRingSegments = 20;
		private const int numberTubeSegments = 10;
		
		private const int numberTriangles = 2*numberRingSegments*numberTubeSegments;
		
		private Device device;
		private VertexBuffer myVertices;
		private Random myRandom = new Random();
		private Matrix worldMatrix = Matrix.Identity;
		private int numberTrianglesToDraw = numberTriangles;
		private bool useZBuffer = true;
		
		public MainClass()
		{
			this.ClientSize = new System.Drawing.Size(640, 480);
			this.Text = "Direct3D Project";
		}
		
		public bool InitializeGraphics()
		{
			try {
				PresentParameters presentParams = new PresentParameters();
				presentParams.Windowed = true;
				presentParams.SwapEffect = SwapEffect.Discard;
				presentParams.AutoDepthStencilFormat = DepthFormat.D24X8;
				presentParams.EnableAutoDepthStencil = true;				
				
				device = new Device(0, DeviceType.Hardware, this, CreateFlags.SoftwareVertexProcessing, presentParams);
				
				device.DeviceLost     += new EventHandler(this.InvalidateDeviceObjects);
				device.DeviceReset    += new EventHandler(this.RestoreDeviceObjects);
				device.Disposing      += new EventHandler(this.DeleteDeviceObjects);
				device.DeviceResizing += new CancelEventHandler(this.EnvironmentResizing);
				
				myVertices = buildTorus(device, 1.0, numberRingSegments, 0.6, numberTubeSegments);

				return true;
			} catch (DirectXException) {
				return false;
			}
		}
		
		protected virtual void InvalidateDeviceObjects(object sender, EventArgs e)
		{
		}
		
		protected virtual void RestoreDeviceObjects(object sender, EventArgs e)
		{
		}
		
		protected virtual void DeleteDeviceObjects(object sender, EventArgs e)
		{
		}
		
		protected virtual void EnvironmentResizing(object sender, CancelEventArgs e)
		{
		}
		
		protected virtual void Render()
		{
			if (device != null) {
				device.Clear(ClearFlags.Target | ClearFlags.ZBuffer, Color.White, 1.0f, 0);
				device.BeginScene();
				
				device.Transform.World = worldMatrix;                            
				device.Transform.View = Matrix.LookAtLH(new Vector3(0.0f, 0.0f, -7.0f), new Vector3(), new Vector3(0.0f, 1.0f, 0.0f));
				device.Transform.Projection = Matrix.PerspectiveFovLH(Geometry.DegreeToRadian(20.0f), ClientSize.Width/(float)ClientSize.Height, 0.1f, 50.0f);
				
				device.RenderState.ZBufferEnable = useZBuffer;
				
				device.SetStreamSource(0, myVertices, 0); 
				device.VertexFormat = CustomVertex.PositionColored.Format; 
				device.RenderState.Lighting = false; 
				device.RenderState.CullMode = Cull.None;
				device.DrawPrimitives(PrimitiveType.TriangleList, 0, numberTrianglesToDraw);
				
				device.EndScene();
				device.Present();
			}
		}
		
		public void Run()
		{
			while (Created) {
				Render();
				Application.DoEvents();
			}
		}
		
		protected override void OnPaint(PaintEventArgs e)
		{
			this.Render();
		}
		
		protected override void OnKeyDown(KeyEventArgs e)
		{
			switch(e.KeyCode)
			{
				case Keys.Escape:
					this.Close(); 
					break;
				case Keys.PageUp:
					if(numberTrianglesToDraw < numberTriangles)
					{
						numberTrianglesToDraw++;
					}
					break;
				case Keys.PageDown:
					if(numberTrianglesToDraw > 1)
					{
						numberTrianglesToDraw--;
					}
					break;
				case Keys.D:
					useZBuffer = !useZBuffer;
					break;
				case Keys.Up:
					worldMatrix = worldMatrix*Matrix.RotationX( 0.1f);
					break;
				case Keys.Down:
					worldMatrix = worldMatrix*Matrix.RotationX(-0.1f);
					break;					
				case Keys.Left:
					worldMatrix = worldMatrix*Matrix.RotationY( 0.1f);
					break;
				case Keys.Right:
					worldMatrix = worldMatrix*Matrix.RotationY(-0.1f);
					break;			
			}
		}
		
		static void Main()
		{
			using (MainClass mainClass = new MainClass()) {
				if (!mainClass.InitializeGraphics()) {
					MessageBox.Show("Error while initializing Direct3D");
					return;
				}
				mainClass.Show();
				mainClass.Run();
			}
		}
		
		private VertexBuffer buildTorus(Device device, double ringRadius, int ringSegments, double tubeRadius, int tubeSegments)
		{
			// Better do this using an additional IndexBuffer, J.L.
			
			VertexBuffer vb = new VertexBuffer(typeof(CustomVertex.PositionColored), 6*ringSegments*tubeSegments, device, Usage.WriteOnly, 0, Pool.Managed);
			CustomVertex.PositionColored[] v = (CustomVertex.PositionColored[]) vb.Lock(0, 0);
			for(int ringIndex = 0; ringIndex < ringSegments; ringIndex++)
			{
				for(int tubeIndex = 0; tubeIndex < tubeSegments; tubeIndex++)
				{
					int baseIndex = 6*(tubeIndex+tubeSegments*ringIndex);
					
					// one quadrangular face
					Vector3 a = torusPoint(ringRadius, ringSegments, ringIndex,
					                       tubeRadius, tubeSegments, tubeIndex);
					Vector3 b = torusPoint(ringRadius, ringSegments, ringIndex + 1,
					                       tubeRadius, tubeSegments, tubeIndex);
					Vector3 c = torusPoint(ringRadius, ringSegments, ringIndex + 1,
					                       tubeRadius, tubeSegments, tubeIndex + 1);					
					Vector3 d = torusPoint(ringRadius, ringSegments, ringIndex,
					                       tubeRadius, tubeSegments, tubeIndex + 1);					
					
					// turn quadrangle into two triangles
					v[baseIndex    ].Position = a;
					v[baseIndex + 1].Position = b;
					v[baseIndex + 2].Position = d;
					v[baseIndex + 3].Position = b;
					v[baseIndex + 4].Position = c;
					v[baseIndex + 5].Position = d;
					
					v[baseIndex    ].Color
						= v[baseIndex + 1].Color
						= v[baseIndex + 2].Color
						= myRandom.Next(0xFFFFFF);
					
					v[baseIndex + 3].Color
						= v[baseIndex + 4].Color
						= v[baseIndex + 5].Color
						= myRandom.Next(0xFFFFFF);					
				}
			}
			vb.Unlock();
			return vb;
		}
		
		private Vector3 torusPoint(double ringRadius, int ringSegments, int ringIndex,
		                            double tubeRadius, int tubeSegments, int tubeIndex)
		{
			double ringAngle = 2.0*Math.PI*ringIndex/ringSegments;
			double tubeAngle = 2.0*Math.PI*tubeIndex/tubeSegments;
			double radialDistance = ringRadius + tubeRadius*Math.Cos(tubeAngle);
			return new Vector3((float)(radialDistance * Math.Cos(ringAngle)),
			                   (float)(tubeRadius * Math.Sin(tubeAngle)),
			                   (float)(radialDistance * Math.Sin(ringAngle))
			                  );
		}
	}
}
